#include "StdAfx.h"
#include "RawSpeed-API.h"
#include "FileWriter.h"

/* 
RawSpeed - RAW file decoder.

Copyright (C) 2009 Klaus Post

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.klauspost.com
*/

using namespace RawSpeed;

int startTime;

static void shuffleShorts(ushort16 *in_array, uint32 elements)
{
  uchar8* a = (uchar8*)in_array;
  uchar8* temp = new uchar8[elements*2];
  for (uint32 i = 0; i < elements; i++)
  {
    temp[i] = a[i*2];
    temp[i+elements] = a[i*2+1];
  }
  memcpy(a, temp, elements*2);
  delete[] temp;
}

static __inline ushort16 ZigZag(short word) {
  return (word >> 15) ^ (word << 1);
}

// Open file and save as tiff
void OpenFile(FileReader f, CameraMetaData *meta) {
  RawDecoder *d = 0;
  FileMap* m = 0;
  try {
        wprintf(L"Opening:%s\n",f.Filename());
    try {
      m = f.readFile();    	
    } catch (FileIOException &e) {
      printf("Could not open image:%s\n", e.what());
      return;
    }
    RawParser t(m);
    d = t.getDecoder();
    d->checkSupport(meta);
    startTime = GetTickCount();

    d->decodeRaw();
//    d->decodeMetaData(meta);
    RawImage r = d->mRaw;
//    r->scaleBlackWhite();

    uint32 time = GetTickCount()-startTime;
    float mpps = (float)r->dim.x * (float)r->dim.y * (float)r->getCpp()  / (1000.0f * (float)time);
    wprintf(L"Decoding %s took: %u ms, %4.2f Mpixel/s\n", f.Filename(), time, mpps);
    for (uint32 i = 0; i < r->errors.size(); i++) {
      printf("Error Encountered:%s", r->errors[i]);
    }
/*    if (!r->isCFA) {
      startTime = GetTickCount();
      MipSaver mip(m, r);
      mip.save(f.Filename());
      time = GetTickCount()-startTime;
      mpps = (float)r->dim.x * (float)r->dim.y * (float)r->getCpp()  / (1000.0f * (float)time);
      wprintf(L"Encoding %s took: %u ms, %4.2f Mpixel/s\n", f.Filename(), time, mpps);
    }*/
    vector<TiffIFD*> data = d->getRootIFD()->getIFDsWithTag(TiffTag(0xc5d8));
    if (data.empty())
      data = d->getRootIFD()->getIFDsWithTag(CFAPATTERN);
    if (data.empty())
      data = d->getRootIFD()->getIFDsWithTag(STRIPOFFSETS);
    
    if (data.empty())
      ThrowRDE("Unable to locate probable RAW data");

    TiffIFD* raw = data[0];
    TiffEntry *offsets = raw->getEntry(STRIPOFFSETS);
    TiffEntry *counts = raw->getEntry(STRIPBYTECOUNTS);
    // Iterate through all slices
    if (!(offsets->isInt() || counts->isInt()))
      ThrowRDE("offset/counts is not int");
    const uint32* off_a = offsets->getIntArray();
    const uint32* count_a = counts->getIntArray();

    for (uint32 s = 0; s < offsets->count; s++) {
      memset(m->getDataWrt(off_a[s]),0, count_a[s]);
    }
    WCHAR filename[65536];
    wsprintf(filename, L"%s-separated", f.Filename());
    CreateDirectoryW(filename,NULL);
    wsprintf(filename, L"%s-separated\\image-shell.dat", f.Filename());
    FileWriter w(filename);      
    w.writeFile(m);
    int width= r->dim.x;
    int height= r->dim.y;
    FileMap imageData(r->dim.area()*2);
    for (int i = 0; i < height; i++)
      memcpy(imageData.getDataWrt(i*width*2),r->getData(0,i),width*2);
    
    wsprintf(filename, L"%s-separated\\image-unmodified.dat", f.Filename());
    w = FileWriter(filename);
    w.writeFile(&imageData);

    FileMap shuffled(r->dim.area()*2);
/*    for (int i = 0; i < height; i++)
      memcpy(shuffled.getDataWrt(i*width*2),r->getData(0,i),width*2);
    shuffleShorts((ushort16*)shuffled.getDataWrt(0), shuffled.getSize()/2);
    wsprintf(filename, L"%s-separated\\image-unmodified-shuffled.dat", f.Filename());
    w = FileWriter(filename);
    w.writeFile(&shuffled);
*/
    FileMap delta(r->dim.area()*2);
    for (int y = 0; y < height; y++) {
      ushort16 *src = (ushort16*)r->getData(0,y);
      ushort16 *dst = (ushort16*)delta.getDataWrt(y*width*2);
      // Save initial values
      dst[0] = src[0];
      dst[1] = src[1];
      for (int x = 2; x < width; x+=2) {
        int delta1 = src[x] - (int)src[x-2];
        int delta2 = src[x+1] - (int)src[x-1];
        dst[x] = ZigZag(delta1);
        dst[x+1] = ZigZag(delta2);
      }
    }
    wsprintf(filename, L"%s-separated\\image-delta-zigzag.dat", f.Filename());
    w = FileWriter(filename);
    w.writeFile(&delta);

    memcpy(shuffled.getDataWrt(0),delta.getData(0),delta.getSize());
    shuffleShorts((ushort16*)shuffled.getDataWrt(0), shuffled.getSize()/2);
    wsprintf(filename, L"%s-separated\\image-delta-zigzag-shuffled.dat", f.Filename());
    w = FileWriter(filename);
    w.writeFile(&shuffled);

  } catch (RawDecoderException &e) {
    wchar_t uni[1024];
    MultiByteToWideChar(CP_ACP, 0, e.what(), -1, uni, 1024);
    //    MessageBox(0,uni, L"RawDecoder Exception",0);
    wprintf(L"Raw Decoder Exception:%s\n",uni);
  }
  if (d) delete d;
  if (m) delete m;

}

int wmain(int argc, _TCHAR* argv[])
{
  if (1) {  // for memory detection
    try {
//          CameraMetaData meta;

      CameraMetaData meta("cameras.xml");

      if (argc > 1)
        OpenFile(FileReader(argv[1]),&meta);
      else
        wprintf(L"You must specify file to split\n");

      /*
      OpenFile(FileReader(L"..\\testimg\\bench\\canon_eos_5d_mark_iii_05.cr2"),&meta);
        OpenFile(FileReader(L"..\\testimg\\bench\\canon_eos_6d_14.cr2"),&meta);
        OpenFile(FileReader(L"..\\testimg\\bench\\canon_eos_m_04.cr2"),&meta);
        OpenFile(FileReader(L"..\\testimg\\bench\\nikon_1_v2_17.nef"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\nikon_d4_10.nef"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\nikon_d5200_14.nef"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\olympus_epm2_16.orf"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\olympus_om_d_e_m5_24.orf"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\olympus_xz2_10.orf"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\panasonic_lumix_dmc_gh3_10.rw2"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\panasonic_lumix_g5_15.rw2"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\pentax_k5_ii_12.dng"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\pentax_q10_19.dng"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\samsung_nx1000_19.srw"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\samsung_nx20_01.srw"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\sony_a55.arw"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\sony_a77_08.arw"),&meta);
          OpenFile(FileReader(L"..\\testimg\\bench\\sony_a99_04.arw"),&meta);
//      OpenFile(FileReader(L"..\\testimg\\5diso\\IMG_8800.CR2"),&meta);
*/

    } catch (CameraMetadataException) {

    }
//    MessageBox(0,L"Finished", L"Finished",0);
#ifdef _USE_GFL_
    gflLibraryExit();
#endif
  } // Dump objects
  _CrtDumpMemoryLeaks();
  return 0;
}

